#!/usr/local/bin/python 

##
#	Honyaku: a biology-driven Bayesian methodology for identifying translation initiation sites in prokaryotes
#	For Pseudomonas aeruginosa
##


from Numeric import *
from string import *
import sys
import statistics

#Ptls = 3462./3424308		#Cenibacterium
#Ptls = 4242./4639675		#E.coli
#Ptls = 858./17774		#Ecoli possible translation start site (over 30aa)
#Ptls = 858./19193		#Ecoli possible translation start site (over 1aa)
#Ptls = 4105./4214630		#Bsubtilis
Ptls = 5567./6264403		#Pseudomonas aeruginosa PAO1


### setup parameters
min_pro_len = 20 #minimum length of searching protein (aa)

# read *.fna file
genome = open("../DB_etc/NC_002516.fna").read()
i = genome.index("\n")
genome = genome[i:]
genome = genome.replace("\n","")


# read overlapping length file
ovlen = open("../overlap_operon/iglen_ud50").readlines()
for i in range(len(ovlen)):
	ovlen[i] = float(ovlen[i])
ovlen = array(ovlen)
max_ovlen = 101
min_ovlen = -40 #range from -40 to +100
x_ovlen = arange(min_ovlen,max_ovlen)
y_ovlen = statistics.pdf(ovlen, x_ovlen, kernel="Gaussian")

# read overlapping length file
ovlen = open("../overlap_operon/iglen_HOH").readlines()
for i in range(len(ovlen)):
	ovlen[i] = float(ovlen[i])
ovlen = array(ovlen)
max_ovlen = 101
min_ovlen = -40 #range from -40 to +100
x_ovlen = arange(min_ovlen,max_ovlen)
y_ovlen_all = statistics.pdf(ovlen, x_ovlen, kernel="Gaussian")

#for checking the overlapping case
ptt = open("../DB_etc/NC_002516.ptt").readlines()
ptt = ptt[3:]
for i in range(len(ptt)):
	ptt[i] = ptt[i].replace("\n","")
	ptt[i] = ptt[i].split("\t")
	ptt[i][0] = ptt[i][0].split("..")

# read gap_len file
dis = open("../distance_dist/gap_len").readlines()
for i in range(len(dis)):
	dis[i] = float(dis[i])
dis = array(dis)
max_dis = 15
if max(dis) > 15:
	max_dis = max(dis)
x = arange(0,max_dis)
y = statistics.pdf(dis, x, kernel="Gaussian")

# read protein length file
plen = open("../pro_len/ratio").readlines()
for i in range(len(plen)):
	plen[i] = float(plen[i])
for i in range(len(plen)):	# mirroring
	plen.append(1+abs(1-plen[i]))

plen = array(plen)
max_data = 1.0
min_data = 0.0
h = 0.01
x_plen = arange(min_data,max_data+h/3,h)	#for making 1.0 data
y_plen = statistics.pdf(plen, x_plen, kernel="Gaussian")
y_plen = y_plen * 2		# mirroring
#y_plen[0]=y_plen[1]=y_plen[2]=y_plen[3]=y_plen[4]=y_plen[5]=y_plen[6]=y_plen[7]=y_plen[8]=y_plen[9]=y_plen[10]=y_plen[11]=y_plen[12]=y_plen[13]=y_plen[14]=y_plen[15]=y_plen[16]=y_plen[17]=y_plen[18]=y_plen[19]

#####################################################################################
def findstart(allseq,ter_pos):
	around_start_seq = []
	TLS_pos = []
	for i in range(len(allseq))[2:]:
		aaseq = allseq[ter_pos-i*3:ter_pos-i*3+3]
		if aaseq in ["ATG","GTG","TTG"]:
			if i > min_pro_len:
				around_start_seq.append(allseq[ter_pos-i*3-50:ter_pos-i*3+20])
				TLS_pos.append(ter_pos-i*3+1)
				
		elif aaseq in ["TAA","TAG","TGA"]:
			break
		elif ter_pos-i*3 < 0:	#for the last gene which doesn't have the next stop codon
			break
	return around_start_seq,TLS_pos

#####################################################################################
def pssmscore(upseq_seq,pssm):
	#seqrch PSSM max score
	total = 0
	for j in range(len(upseq_seq)-1):
		score = 0
		flag = 0
		second = -1
		for a in "ACGT":
			second += 1
			first = -1
			for b in "ACGT":
				first += 1
				if upseq_seq[j:j+2] == a+b:
					score = pssm[j][second*4+first]
					total = total + score
					flag = 1
					break
			if flag == 1:
				break
	return total
######################################################################################
def compl(seq):
	complseq = seq[-1::-1]
#	for c in range(len(seq)):
#		complseq = complseq + seq[-c-1]
	complseq = complseq.replace("A","t")
	complseq = complseq.replace("T","a")
	complseq = complseq.replace("G","c")
	complseq = complseq.replace("C","g")
	complseq = swapcase(complseq)
	
	return complseq
#####################################################################################
wm = open("../make_PSSM/align_matrix_1st/align_dw20.txt").readlines()
pssm20 = []
for wmline in wm:
	wmline = wmline.replace("\n","")
	wmline = wmline.split("\t")
	wmline = map(float, wmline)
	pssm20.append(wmline)
pssm20 = array(pssm20)
pssm20 = pssm20[3:]		#remove ATC start codon
	
wm = open("../make_PSSM/align_matrix_1st/align_up30.txt").readlines()
pssm30 = []
for wmline in wm:
	wmline = wmline.replace("\n","")
	wmline = wmline.split("\t")
	wmline = map(float, wmline)
	pssm30.append(wmline)
pssm30 = array(pssm30)

########################################################################################
def prob_stcodon(stc):
	if stc == "ATG":
		return 0.8847
	elif stc == "GTG":
		return 0.0922
	elif stc == "TTG":
		return 0.0231
	else:
		print "ERROR: start codon is not much with ATG, GTG, or TTG!",stc
########################################################################################
def cal_score(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,direc,next_direc):

	printout = []
	TLS_end_pos = int(TLS_end_pos)
	
	for ino in range(len(around_start_seq)):
		sc_wm20 = pssmscore(around_start_seq[ino][-21+3:],pssm20)
		max_sc = -1000
		sc_stcodon = prob_stcodon(around_start_seq[ino][-20:-17])
		pro_len = (abs(TLS_pos[ino]-TLS_end_pos)+1)/3.
		longest = (abs(TLS_pos[-1]-TLS_end_pos)+1)/3.
		
		if y_plen[int(round(pro_len/longest*100))] != 0 :
			y_plen_val = y_plen[int(round(pro_len/longest*100))]
			sc_pro_len = log(y_plen_val)
		else:
			sc_pro_len = -100
			y_plen_val = 0
		
		for gap in range(15):
			sc_dis = log(y[gap])	#only seqrch from 0 to 15bp
			sc_wm30 = pssmscore(around_start_seq[ino][-20-31-gap:-20-gap],pssm30)
			sc_total = sc_wm30 + sc_dis + sc_wm20 + log(sc_stcodon) + sc_pro_len
			if max_sc < sc_total:
				max_wm30 = sc_wm30
				max_wm20 = sc_wm20
				max_sc = sc_total
				max_seq = around_start_seq[ino][-20-31-gap:]
				max_pos = int(TLS_pos[ino])
				max_pro_len = pro_len
				max_sc_dis = sc_dis
				max_start_codon = log(sc_stcodon)
				max_sc_pro_len = sc_pro_len
		
		# In case there is a math range error under -745 (exp(-745)) 
		if max_sc < -745:       
			max_sc = -745
		bayes = exp(max_sc)*Ptls/(exp(max_sc)*Ptls+(1./40)*(1-Ptls))
		printout.append([gene_name,max_sc,bayes,int(max_pos),max_pro_len,max_seq,max_wm30,max_wm20,max_sc_dis,max_start_codon,max_sc_pro_len])

	printout.sort(lambda x, y: cmp(x[1],y[1]))
	fg_ov = 0
	if(direc == next_direc):
		if (direc == "-"):
			if (next_stop - printout[-1][3] <= 50):
				printout = cal_score_ov(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,direc,next_direc,y_ovlen)
				fg_ov = 1
		else:
			if (printout[-1][3] - next_stop <= 50):
				printout = cal_score_ov(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,direc,next_direc,y_ovlen)
				fg_ov = 1
	else:
		if (direc == "-"):
			if (next_stop - printout[-1][3] <= 50):
				printout = cal_score_ov(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,direc,next_direc,y_ovlen_all)
				fg_ov = 1
		else:
			if (printout[-1][3] - next_stop <= 50):
				printout = cal_score_ov(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,direc,next_direc,y_ovlen_all)
				fg_ov = 1
	
	return(printout,fg_ov)

########################################################################################
def cal_score_ov(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,direc,next_direc,y_ovlen):

	printout = []
	TLS_end_pos = int(TLS_end_pos)
	
	for ino in range(len(around_start_seq)):
		sc_wm20 = pssmscore(around_start_seq[ino][-21+3:],pssm20)
		max_sc = -1000
		sc_next_stop = 0
		if (direc == "-"):
			if (next_stop-TLS_pos[ino] < -35):
				sc_next_stop = log(y_ovlen[0])
			elif(next_stop-TLS_pos[ino] >= 100):
				sc_next_stop = log(y_ovlen[-1])
			else:
				sc_next_stop = log(y_ovlen[next_stop-TLS_pos[ino]-1+39])	#+39:y_ovlen[-40:+100]
		else:
			if (TLS_pos[ino]-next_stop < -35):
				sc_next_stop = log(y_ovlen[0])
			elif(TLS_pos[ino]-next_stop >= 100):
				sc_next_stop = log(y_ovlen[-1])
			else:
				sc_next_stop = log(y_ovlen[TLS_pos[ino]-next_stop-1+39])
		
		sc_stcodon = prob_stcodon(around_start_seq[ino][-20:-17])
		pro_len = (abs(TLS_pos[ino]-TLS_end_pos)+1)/3.
		longest = (abs(TLS_pos[-1]-TLS_end_pos)+1)/3.

		if y_plen[int(round(pro_len/longest*100))] != 0 :
			y_plen_val = y_plen[int(round(pro_len/longest*100))]
			sc_pro_len = log(y_plen_val)
		else:
			sc_pro_len = -100
			y_plen_val = 0

		for gap in range(15):
			sc_dis = log(y[gap])	#only seqrch from 0 to 15bp
			sc_wm30 = pssmscore(around_start_seq[ino][-20-31-gap:-20-gap],pssm30)
			sc_total = sc_wm30 + sc_dis + sc_wm20 + log(sc_stcodon) + sc_pro_len + sc_next_stop
			if max_sc < sc_total:
				max_wm30 = sc_wm30
				max_wm20 = sc_wm20
				max_sc = sc_total
				max_seq = around_start_seq[ino][-20-31-gap:]
				max_pos = TLS_pos[ino]
				max_pro_len = pro_len
				max_sc_dis = sc_dis
				max_start_codon = log(sc_stcodon)
				max_sc_pro_len = sc_pro_len
				max_sc_next_stop = sc_next_stop
		# In case there is a math range error under -745 (exp(-745)) 
		if max_sc < -745:       
			max_sc = -745
		bayes = exp(max_sc)*Ptls/(exp(max_sc)*Ptls+(1./40)*(1-Ptls))
		
		printout.append([gene_name,max_sc,bayes,int(max_pos),max_pro_len,max_seq,max_wm30,max_wm20,max_sc_dis,max_start_codon,max_sc_pro_len,max_sc_next_stop,next_stop,direc])
		
	return(printout)

########################################################################################
def ck_next_stop(ptt,ptt_line):
	
	for l in range(len(ptt)):
		next_stop = 0
		next_direc = ""
		if(int(ptt_line[0][0]) == int(ptt[l][0][0])):
			if(ptt[l][1] == "-"):
				# fro the last gene without upstream seq in *.fna file
				if l == len(ptt)-1:
					next_stop = int(ptt[0][0][0])+len(genome)
					next_direc = ptt[0][1]
				else:
					next_stop = int(ptt[l+1][0][0])
					next_direc = ptt[l+1][1]
				break
			else:
				next_stop = int(ptt[l-1][0][1])
				next_direc = ptt[l-1][1]
				break
	return(next_stop,next_direc)
########################################################################################
def printall(printout):
	fg_print = 0
	
	if len(printout) == 1:
		pretty_print(printout[-1])
	else:
		for pline in printout:
			pretty_print(pline)

########################################################################################
def parse(printout,fg_ov):	#considering two Met locate next to each other..
	fg_print = 0
	
	if len(printout) == 1:
		pretty_print(printout[-1])
	else:
		printout.sort(lambda x, y: cmp(x[1],y[1]))
		if(abs(printout[-1][4]-printout[-2][4]) <= 3):
			if(printout[-1][1]-printout[-2][1] < 3):
				if(printout[-1][5][0:15] == printout[-2][5][0:15]): 
					len1 = abs(57-len(printout[-1][5]))
					len2 = abs(57-len(printout[-2][5]))
					if (len1 > len2):
						pretty_print(printout[-2])
						fg_print = 1
		if(fg_print == 0):
			pretty_print(printout[-1])

########################################################################################
def pretty_print(pprint):
	print "%s\t%.1f\t%.3f\t%d\t%d\t%s\t%.1f\t%.1f\t%.1f\t%.1f\t%.1f" %(pprint[0],pprint[1],pprint[2],pprint[3],pprint[4],pprint[5],pprint[6],pprint[7],pprint[8],pprint[9],pprint[10])
			
########################################################################################
print "gene\tscore\tBaysian_probability\tgenomic_location\tlength(aa)\tSequence\tWM_SD\tWM_ATG\tGap_length\tStart_codon\tProtein_length"
genome_compl = compl(genome)

for k in range(len(ptt)):
	#fro the last gene without upstream seq in *.fna file
	if k == len(ptt)-1:
		genome_compl = genome_compl*2
	if ptt[k][1] == "-":
		TLS_end_pos = int(ptt[k][0][0])
		TLS_start_pos = int(ptt[k][0][1])
		around_start_seq,TLS_pos = findstart(genome_compl,len(genome_compl)-TLS_end_pos+1)
		for kk in range(len(TLS_pos)):
			TLS_pos[kk] = len(genome_compl)-TLS_pos[kk]+1
	else:
		TLS_end_pos = int(ptt[k][0][1])
		TLS_start_pos = int(ptt[k][0][0])
		around_start_seq,TLS_pos = findstart(genome,TLS_end_pos)
	
	gene_name = ptt[k][5]
	#for overlaping length inside a operon
	next_stop,next_direc = ck_next_stop(ptt,ptt[k])
	if around_start_seq != []:
		printout,fg_ov = cal_score(around_start_seq,TLS_pos,gene_name,TLS_end_pos,next_stop,ptt[k][1],next_direc)
		
		# To print every candidate
#		printall(printout)
		# To print only the highest candidate
 		parse(printout,fg_ov)
